## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) import eudplib.string.tblprint;
from eudplib.string import tblprint
# (Line 2) const inputData = py_bytes(1000);
inputData = _CGFW(lambda: [bytes(1000)], 1)[0]
# (Line 3) tblprint._AddStatText(inputData);
tblprint._AddStatText(inputData)
# (Line 5) const expected_result = py_str("abcdeArmo\xE2\x80\x89\0");
expected_result = _CGFW(lambda: [str("abcdeArmo\xE2\x80\x89\0")], 1)[0]
# (Line 6) function test_stattext() {
@EUDFunc
def f_test_stattext():
    # (Line 7) const armo = Db("Armo");
    armo = Db("Armo")
    # (Line 8) const stattext = GetTBLAddr(1);
    stattext = GetTBLAddr(1)
    # (Line 9) settblf(1, 0, "abcde{:s}", armo, encoding="UTF-8");
    f_settblf(1, 0, "abcde{:s}", armo, encoding="UTF-8")
    # (Line 10) const ret = py_list();
    ret = list()
    # (Line 11) const br = EUDByteReader();
    br = EUDByteReader()
    # (Line 12) br.seekoffset(stattext);
    br.seekoffset(stattext)
    # (Line 13) foreach(char : expected_result) {
    for char in expected_result:
        # (Line 14) ret.append(br.readbyte());
        ret.append(br.readbyte())
        # (Line 15) }
        # (Line 16) return List2Assignable(ret);

    EUDReturn(List2Assignable(ret))
    # (Line 17) }



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
import functools

from ... import core as c
from ...memio.rwcommon import lv

_signflag = lv


def f_div_towards_zero(
    a, b, **kwargs
) -> tuple[c.EUDVariable, c.EUDVariable] | tuple[int, int]:
    """Calculates the quotient and remainder of (a ÷ b),
    rounding the quotient towards zero.

    Calculate signed division, unlike unsigned division `f_div(a, b)`.
    Consistent with C-like languages including JavaScript.
    """
    if c.IsEUDVariable(b):
        if not hasattr(f_div_towards_zero, "_eudf"):

            @c.EUDFunc
            def div_towards_zero_eud(a, b):
                return _div_towards_zero(a, b)

            setattr(f_div_towards_zero, "_eudf", div_towards_zero_eud)

        return getattr(f_div_towards_zero, "_eudf")(a, b, **kwargs)

    elif c.IsEUDVariable(a):
        return _div_towards_zero_const(b)(a, **kwargs)

    return int(a / b), a - int(a / b) * b


@functools.cache
def _div_towards_zero_const(b: int):
    @c.EUDFunc
    def f(a):
        return _div_towards_zero(a, b)

    return f


def _div_towards_zero(a, b):
    global _signflag
    _signflag << (0 if c.IsEUDVariable(b) else 2 * bool(b < 0))

    c.RawTrigger(  # a is negative
        conditions=a >= 1 << 31,
        actions=(
            _signflag.AddNumber(1),
            a.ineg(action=True),
        ),
    )
    if c.IsEUDVariable(b):
        c.RawTrigger(  # b is negative
            conditions=b >= 1 << 31,
            actions=(
                _signflag.AddNumber(2),
                b.ineg(action=True),
            ),
        )
    else:
        b = abs(b)

    quotient, modulo = c.f_div(a, b)

    # when only one of divider or dividend is negative, quotient is negative
    c.RawTrigger(
        conditions=(1 <= _signflag, _signflag <= 2),
        actions=quotient.ineg(action=True),
    )
    # when dividend is negative, modulo is negative
    c.RawTrigger(
        conditions=_signflag.ExactlyX(1, 1), actions=modulo.ineg(action=True)
    )

    return quotient, modulo


def f_div_floor(
    a, b, **kwargs
) -> tuple[c.EUDVariable, c.EUDVariable] | tuple[int, int]:
    """Calculates the quotient and remainder of (a ÷ b),
    rounding the quotient towards negative infinity.

    Calculate signed division, unlike unsigned division `f_div(a, b)`.
    Consistent with mathematical modulo.
    """
    if c.IsEUDVariable(b):
        if not hasattr(f_div_floor, "_eudf"):

            @c.EUDFunc
            def div_floor_eud(a, b):
                return _div_floor(a, b)

            setattr(f_div_floor, "_eudf", div_floor_eud)

        return getattr(f_div_floor, "_eudf")(a, b, **kwargs)

    elif c.IsEUDVariable(a):
        return _div_floor_const(b)(a, **kwargs)

    return divmod(a, b)


@functools.cache
def _div_floor_const(b: int):
    @c.EUDFunc
    def f(a):
        return _div_floor(a, b)

    return f


def _div_floor(a, b):
    global _signflag
    _signflag << (0 if c.IsEUDVariable(b) else 2 * bool(b < 0))

    c.RawTrigger(  # a is negative
        conditions=a >= 1 << 31,
        actions=(
            _signflag.AddNumber(1),
            a.ineg(action=True),
        ),
    )
    if c.IsEUDVariable(b):
        c.RawTrigger(  # b is negative
            conditions=b >= 1 << 31,
            actions=(
                _signflag.AddNumber(2),
                b.ineg(action=True),
            ),
        )
    else:
        b = abs(b)

    quotient, modulo = c.f_div(a, b)

    check, ontrue, onfalse = (c.Forward() for _ in range(3))
    if c.IsEUDVariable(b):
        check << c.RawTrigger(  # modulo << -modulo + b
            nextptr=onfalse,
            conditions=(1 <= _signflag, _signflag <= 2, modulo >= 1),
            actions=[
                c.SetNextPtr(check, b.GetVTable()),
                c.SetNextPtr(b.GetVTable(), ontrue),
                quotient.AddNumber(1),
                modulo.ineg(action=True),
                b.QueueAddTo(modulo),
            ],
        )
    else:
        check << c.RawTrigger(  # modulo << -modulo + b
            nextptr=onfalse,
            conditions=(1 <= _signflag, _signflag <= 2, modulo >= 1),
            actions=[
                c.SetNextPtr(check, ontrue),
                quotient.AddNumber(1),
                modulo.ineg(action=True),
                modulo.AddNumber(b),
            ],
        )

    ontrue << c.RawTrigger(actions=c.SetNextPtr(check, onfalse))
    # when only one of divider or dividend is negative, quotient is negative
    onfalse << c.RawTrigger(
        conditions=(1 <= _signflag, _signflag <= 2),
        actions=quotient.ineg(action=True),
    )
    # when divider is negative, remainder is negative
    c.RawTrigger(
        conditions=_signflag.ExactlyX(2, 2), actions=modulo.ineg(action=True)
    )

    return quotient, modulo


def f_div_euclid(
    a, b, **kwargs
) -> tuple[c.EUDVariable, c.EUDVariable] | tuple[int, int]:
    """Calculates the quotient and remainder of Euclidean division of a by b.

    Calculate signed division, unlike unsigned division `f_div(a, b)`.
    This computes the quotient such that `a = quotient * b + remainder`,
    and `0 <= r < abs(b)`.

    In other words, the result is a ÷ b rounded to the quotient
    such that `a >= quotient * b`.
    If `a > 0`, this is equal to round towards zero; if `a < 0`,
    this is equal to round towards +/- infinity (away from zero).
    """
    if c.IsEUDVariable(b):
        if not hasattr(f_div_euclid, "_eudf"):

            @c.EUDFunc
            def div_euclid_eud(a, b):
                return _div_euclid(a, b)

            setattr(f_div_euclid, "_eudf", div_euclid_eud)

        return getattr(f_div_euclid, "_eudf")(a, b, **kwargs)

    elif c.IsEUDVariable(a):
        return _div_euclid_const(b)(a, **kwargs)

    r = a - int(a / b) * b
    r = r + abs(b) if r < 0 else r
    return (a - r) // b, r


@functools.cache
def _div_euclid_const(b: int):
    @c.EUDFunc
    def f(a):
        return _div_euclid(a, b)

    return f


def _div_euclid(a, b):
    global _signflag
    _signflag << (0 if c.IsEUDVariable(b) else 2 * bool(b < 0))

    c.RawTrigger(  # a is negative
        conditions=a >= 1 << 31,
        actions=(
            _signflag.AddNumber(1),
            a.ineg(action=True),
        ),
    )
    if c.IsEUDVariable(b):
        c.RawTrigger(  # b is negative
            conditions=b >= 1 << 31,
            actions=(
                _signflag.AddNumber(2),
                b.ineg(action=True),
            ),
        )
    else:
        b = abs(b)

    quotient, modulo = c.f_div(a, b)

    check, ontrue, onfalse = (c.Forward() for _ in range(3))
    if c.IsEUDVariable(b):
        check << c.RawTrigger(  # modulo << -modulo + b
            nextptr=onfalse,
            conditions=(_signflag.ExactlyX(1, 1), modulo >= 1),
            actions=[
                c.SetNextPtr(check, b.GetVTable()),
                c.SetNextPtr(b.GetVTable(), ontrue),
                quotient.AddNumber(1),
                modulo.ineg(action=True),
                b.QueueAddTo(modulo),
            ],
        )
    else:
        check << c.RawTrigger(  # modulo << -modulo + b
            nextptr=onfalse,
            conditions=(_signflag.ExactlyX(1, 1), modulo >= 1),
            actions=[
                c.SetNextPtr(check, ontrue),
                quotient.AddNumber(1),
                modulo.ineg(action=True),
                modulo.AddNumber(b),
            ],
        )

    ontrue << c.RawTrigger(actions=c.SetNextPtr(check, onfalse))
    # when only one of divider or dividend is negative, quotient is negative
    onfalse << c.RawTrigger(
        conditions=(1 <= _signflag, _signflag <= 2),
        actions=quotient.ineg(action=True),
    )

    return quotient, modulo



from __future__ import annotations

import argparse

import nox

nox.needs_version = ">=2024.3.2"
nox.options.sessions = ["lint", "tests", "tests_packaging"]
nox.options.default_venv_backend = "uv|virtualenv"


@nox.session(reuse_venv=True)
def lint(session: nox.Session) -> None:
    """
    Lint the codebase (except for clang-format/tidy).
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "-a", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """
    Run the tests (requires a compiler).
    """
    tmpdir = session.create_tmp()
    session.install("cmake")
    session.install("-r", "tests/requirements.txt")
    session.run(
        "cmake",
        "-S.",
        f"-B{tmpdir}",
        "-DPYBIND11_WERROR=ON",
        "-DDOWNLOAD_CATCH=ON",
        "-DDOWNLOAD_EIGEN=ON",
        *session.posargs,
    )
    session.run("cmake", "--build", tmpdir)
    session.run("cmake", "--build", tmpdir, "--config=Release", "--target", "check")


@nox.session
def tests_packaging(session: nox.Session) -> None:
    """
    Run the packaging tests.
    """

    session.install("-r", "tests/requirements.txt", "pip")
    session.run("pytest", "tests/extra_python_package", *session.posargs)


@nox.session(reuse_venv=True)
def docs(session: nox.Session) -> None:
    """
    Build the docs. Pass --non-interactive to avoid serving.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b", dest="builder", default="html", help="Build target (default: html)"
    )
    args, posargs = parser.parse_known_args(session.posargs)
    serve = args.builder == "html" and session.interactive

    extra_installs = ["sphinx-autobuild"] if serve else []
    session.install("-r", "docs/requirements.txt", *extra_installs)
    session.chdir("docs")

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        ".",
        f"_build/{args.builder}",
        *posargs,
    )

    if serve:
        session.run(
            "sphinx-autobuild", "--open-browser", "--ignore=.build", *shared_args
        )
    else:
        session.run("sphinx-build", "--keep-going", *shared_args)


@nox.session(reuse_venv=True)
def make_changelog(session: nox.Session) -> None:
    """
    Inspect the closed issues and make entries for a changelog.
    """
    session.install("ghapi", "rich")
    session.run("python", "tools/make_changelog.py")


@nox.session(reuse_venv=True)
def build(session: nox.Session) -> None:
    """
    Build SDists and wheels.
    """

    session.install("build")
    session.log("Building normal files")
    session.run("python", "-m", "build", *session.posargs)
    session.log("Building pybind11-global files (PYBIND11_GLOBAL_SDIST=1)")
    session.run(
        "python", "-m", "build", *session.posargs, env={"PYBIND11_GLOBAL_SDIST": "1"}
    )



# Copyright 2014-2019 by trgk, Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..localize import _
from ..memio.rwcommon import br1, bw1
from .dbstr import DBStringData
from .strcommon import _obfus, epd2s, hptr, ptr2s


def _add_str(dst, *, operations=[], _f=[]):
    try:
        start, end = _f[0], _f[1]
    except IndexError:
        c.PushTriggerScope()
        start = c.NextTrigger()
        jump, write, end = c.Forward(), c.Forward(), c.Forward()
        if cs.EUDInfLoop()():
            b = br1.readbyte()
            jump << c.RawTrigger(
                conditions=b.Exactly(0), actions=c.SetNextPtr(jump, end)
            )
            write << c.NextTrigger()
            bw1.writebyte(b)
            c.RawTrigger(actions=c.SetMemory(end + 380, c.Add, 1))
        cs.EUDEndInfLoop()

        end << c.RawTrigger(
            nextptr=0,
            actions=[
                c.SetMemoryEPD(0, c.Add, 0),
                c.SetMemory(end + 348, c.SetTo, 0),
                c.SetNextPtr(jump, write),
            ],
        )
        c.PopTriggerScope()

        _f.extend((start, end))

    if isinstance(dst, c.EUDVariable):
        dst = ut.EPD(dst.getValueAddr())
    nptr = c.Forward()
    _operations = [
        (ut.EPD(end) + 1, c.SetTo, nptr),
        (ut.EPD(end) + 86, c.SetTo, dst),
    ]
    if operations:
        _operations.extend(operations)
    c.NonSeqCompute(_operations)
    c.SetNextTrigger(start)
    nptr << c.NextTrigger()


@c.EUDFunc
def f_dbstr_addstr(dst, src):
    """Print string as string to dst. Same as strcpy except of return value.

    :param dst: Destination address (Not EPD player)
    :param src: Source address (Not EPD player)

    :returns: dst + strlen(src)
    """
    bw1.seekoffset(dst)
    br1.seekoffset(src)
    _add_str(dst)
    bw1.writebyte(0)

    return dst


@c.EUDFunc
def f_dbstr_addstr_epd(dst, epd):
    """Print string as string to dst. Same as strcpy except of return value.

    :param dst: Destination address (Not EPD player)
    :param epd: Source EPD player

    :returns: dst + strlen_epd(epd)
    """
    bw1.seekoffset(dst)
    br1.seekepd(epd)
    _add_str(dst)
    bw1.writebyte(0)

    return dst


def _add_dw(dst, number, *, operations=[], _f=[]):
    try:
        start, end, number_addr = _f[0], _f[1], _f[2]
    except IndexError:
        c.PushTriggerScope()
        start = c.NextTrigger()
        _number = c.EUDVariable()
        number_addr = ut.EPD(_number.getValueAddr())
        end = c.Forward()

        skipper = [c.Forward() for _ in range(9)]
        ch = [0] * 10

        # Get digits
        for i in range(10):
            _number, ch[i] = c.f_div(_number, 10)
            ch[i] += b"0"[0]
            if i != 9:
                cs.EUDJumpIf(_number == 0, skipper[i])

        # print digits
        for i in range(9, -1, -1):
            if i != 9:
                skipper[i] << c.NextTrigger()
            bw1.writebyte(ch[i])
            c.RawTrigger(actions=c.SetMemory(end + 348, c.Add, 1))

        end << c.RawTrigger(
            nextptr=0,
            actions=[
                c.SetMemoryEPD(0, c.Add, 0),
                c.SetMemory(end + 348, c.SetTo, 0),
            ],
        )
        c.PopTriggerScope()

        _f.extend((start, end, number_addr))

    if isinstance(dst, c.EUDVariable):
        dst = ut.EPD(dst.getValueAddr())
    nptr = c.Forward()
    _operations = [
        (ut.EPD(end) + 1, c.SetTo, nptr),
        (ut.EPD(end) + 86, c.SetTo, dst),
        (number_addr, c.SetTo, number),
    ]
    if operations:
        _operations.extend(operations)
    c.NonSeqCompute(_operations)
    c.SetNextTrigger(start)
    nptr << c.NextTrigger()


@c.EUDFunc
def f_dbstr_adddw(dst, number):
    """Print number as string to dst.

    :param dst: Destination address (Not EPD player)
    :param number: DWORD to print

    :returns: dst + strlen(itoa(number))
    """
    bw1.seekoffset(dst)
    _add_dw(dst, number)
    bw1.writebyte(0)
    return dst


def _add_ptr(dst, number, *, operations=[], _f=[]):
    try:
        start, end, number_addr = _f[0], _f[1], _f[2]
    except IndexError:
        c.PushTriggerScope()
        start = c.NextTrigger()
        _number = c.EUDVariable()
        number_addr = ut.EPD(_number.getValueAddr())
        end = c.Forward()
        ch = [0] * 8

        # Get digits
        for i in range(8):
            _number, ch[i] = c.f_div(_number, 16)
        c.RawTrigger(actions=[c.AddNumber(b"0"[0]) for c in ch])

        # print digits
        for i in range(7, -1, -1):
            c.RawTrigger(
                conditions=ch[i].AtLeast(b"0"[0] + 10),
                actions=ch[i].AddNumber(b"A"[0] - b"0"[0] - 10),
            )
            bw1.writebyte(ch[i])

        end << c.RawTrigger(nextptr=0)
        c.PopTriggerScope()

        _f.extend((start, end, number_addr))

    nptr = c.Forward()
    _operations = [
        (dst, c.Add, 8),
        (ut.EPD(end) + 1, c.SetTo, nptr),
        (number_addr, c.SetTo, number),
    ]
    if operations:
        _operations.extend(operations)
    c.NonSeqCompute(_operations)
    c.SetNextTrigger(start)
    nptr << c.NextTrigger()


@c.EUDFunc
def f_dbstr_addptr(dst, number):
    """Print number as string to dst.

    :param dst: Destination address (Not EPD player)
    :param number: DWORD to print

    :returns: dst + strlen(itoa(number))
    """
    bw1.seekoffset(dst)
    _add_ptr(dst, number)
    bw1.writebyte(0)
    return dst


def f_dbstr_print(dst, *args, EOS=True, encoding="UTF-8"):  # noqa: N803
    """Print multiple string / number to dst.

    :param dst: Destination address (Not EPD player)
    :param args: Things to print

    """
    ret = c.EUDVariable()
    ret << dst
    bw1.seekoffset(dst)
    static_len = 0

    for arg in ut.FlattenIter(args):
        try:
            arg = arg.fmt()
        except AttributeError:
            pass
        x = ut.unProxy(arg)
        if isinstance(x, str):
            x = x.encode(encoding)
        elif isinstance(x, int):
            # int and c.EUDVariable should act the same if possible.
            # EUDVariable has a value of 32bit unsigned integer.
            # So we adjust arg to be in the same range.
            x = str(x & 0xFFFFFFFF).encode(encoding)
        if isinstance(x, bytes):
            if x == b"":
                continue
            _obfus.dbstr_print(x)
            static_len += len(x)
        elif isinstance(x, c.Db | DBStringData):
            _add_str(ret, operations=br1.seekepd(ut.EPD(x), operation=True))
        elif isinstance(x, ptr2s):
            if c.IsEUDVariable(x._value):
                br1.seekoffset(x._value)
                _add_str(ret)
            else:
                _add_str(ret, operations=br1.seekoffset(x._value, operation=True))
        elif isinstance(x, epd2s):
            _add_str(ret, operations=br1.seekepd(x._value, operation=True))
        elif isinstance(x, hptr):
            _add_ptr(ret, x._value)
        elif isinstance(x, c.EUDVariable | c.ConstExpr):
            _add_dw(ret, x)
        else:
            e = _("Object with unknown parameter type {} given to {}")
            raise ut.EPError(e.format(arg, "f_dbstr_print"))

    if static_len:
        ret += static_len

    if EOS:
        bw1.writebyte(0)

    return ret



from helper import *


@TestInstance
def test_strbuffer():
    test_assert("LocalLocale test", [LocalLocale == "koUS"])
    foo = Db(4)
    f_sprintf(foo, "{}", LocalLocale)
    test_equality("LocalLocale print test", b2i4(b"koUS"), f_dwread(foo))

    origcp = f_getcurpl()

    s1 = StringBuffer(1023)

    a = EUDVariable(4006547881)
    s1.insert(0, a)

    b = EUDVariable(0xB91083F5)
    s1.append(hptr(b))

    s1.append(347, "test.")
    EP_SetRValueStrictMode(False)
    s1.delete(1, 1)
    EP_SetRValueStrictMode(True)

    userid = f_getuserplayerid()
    DoActions(SetCurrentPlayer(userid), SetMemory(0x640B58, SetTo, 0))
    s1.Display()

    test_assert(
        "StringBuffer test",
        [
            Memory(0x640B60, Exactly, b2i4(b"4006")),
            Memory(0x640B64, Exactly, b2i4(b"\r\r\r\r")),
            Memory(0x640B68, Exactly, b2i4(b"81\r\r")),
            Memory(0x640B6C, Exactly, b2i4(b"B910")),
            Memory(0x640B70, Exactly, b2i4(b"83F5")),
            Memory(0x640B74, Exactly, b2i4(b"347\r")),
            Memory(0x640B78, Exactly, b2i4(b"test")),
            Memory(0x640B7C, Exactly, b2i4(b".\r\r\r")),
            MemoryX(0x640B80, Exactly, 0, 0xFF),
        ],
    )

    s2 = StringBuffer(1023)
    s2addr = GetMapStringAddr(s2.StringIndex)

    test_assert(
        "StringBuffer uniqueness test",
        [
            s1.StringIndex != s2.StringIndex,
            GetMapStringAddr(s1.StringIndex) != s2addr,
        ],
    )

    f_dwwrite(0x640B58, 6)
    s2.print(PName(userid), "님 안녕하세요")
    username = 0x57EEEB + 36 * userid
    unamelen = f_strlen(username)

    test_equality(
        "PName test",
        [
            f_memcmp(s2addr, username, unamelen),
            f_memcmp(0x640B60 + 218 * 6, username, unamelen),
        ],
        [0, 0],
    )

    cmp_db = Db("SetPNameTest:\x07 테스트입니다.\r\r")  # 36
    f_dbstr_print(0x64107C, ptr2s(username), u2utf8(":\x07 테스트입니다."))
    SetPName(userid, "SetPNameTest")
    test_equality("SetPName test", f_memcmp(0x64107C, cmp_db, 40), 0)

    name_db = Db(25)
    name_var = EUDVariable(name_db)
    f_dbstr_print(name_db, ptr2s(username))
    ret = EUDVariable(0)

    RawTrigger(conditions=IsPName(P1, "Armo"), actions=ret.AddNumber(1))
    RawTrigger(conditions=IsPName(P1, "trgk"), actions=ret.AddNumber(16))
    test_equality("IsPName test1", ret, 0b1)

    Trigger(conditions=IsPName(userid, "Armo"), actions=ret.AddNumber(2))
    Trigger(conditions=IsPName(userid, "trgk"), actions=ret.AddNumber(16))
    test_equality("IsPName test2", ret, 0b11)

    Trigger(conditions=IsPName(P1, name_var), actions=ret.AddNumber(4))
    Trigger(conditions=IsPName(P1, Db("trgk")), actions=ret.AddNumber(16))
    test_equality("IsPName test3", ret, 0b111)

    Trigger(conditions=IsPName(userid, name_var), actions=ret.AddNumber(8))
    Trigger(
        conditions=IsPName(userid, EUDVariable(Db("trgk"))),
        actions=ret.AddNumber(16),
    )
    test_equality("IsPName test4", ret, 0b1111)

    v3 = EUDVariable(s2)
    s3 = StringBuffer.cast(v3)
    f_dwwrite(0x640B58, 8)
    s3.print("dpdkfah!")
    test_equality(
        "StringBuffer.cast(var) test",
        [f_bread(0x641230 + i) for i in range(len(b"dpdkfah!\0"))],
        b"dpdkfah!\0",
    )

    f_dwwrite(0x640B58, 0)
    f_setcurpl(userid)
    if EUDIfNot()(True):
        neverstr = StringBuffer()
        a = Db("Hi")
        neverstr.printf("{:t}", EPD(a))
    if EUDElse()():
        b = Db("Never say never")
        neverstr.printf("{:t}", EPD(b))
    EUDEndIf()
    test_equality(
        "StringBuffer branch test",
        [f_bread(0x640B60 + i) for i in range(len("Never say never\r\0"))],
        b"Never say never\r\0",
    )

    f_setcurpl(userid)
    hammo = StringBuffer.cast(EUDVariable(StringBuffer()))
    for i in range(11):
        f_dwwrite(0x640B58, 0)
        hammo.insert(0)
        hammo.appendf("\x13hammo")
        hammo.DisplayAt(i)
        chatptr = f_dwread(0x640B58)
        test_equality(
            f"StringBuffer.DisplayAt({i}) test",
            [f_bread(0x640B60 + 218 * i + x) for x in range(len("\x13hammo\r\r\0"))],
            b"\x13hammo\r\r\0",
        )
        test_equality(f"StringBuffer.DisplayAt({i}) chatptr test", chatptr, 0)

    s2.fadeIn("abc", line=10)
    s2.fadeIn("abc", line=EUDVariable(-1))
    s2.fadeOut("abc", line=10)
    s2.fadeOut("abc", line=EUDVariable(-1))
    s2.tagprint("dpdkfah")
    s2.tagprint("dpdkfah", line=EUDVariable(-1))
    SetPName(userid, "TestPName{:c}{:n}", userid, userid)
    SetPName(P1, "TestTest")
    SetPName(P2, "TestTest")
    SetPName(P4, "TestTest")
    SetPNamef(userid, "SetPNameTest{}", userid)

    f_setcurpl(origcp)



import datetime
import os
import subprocess
import time

curTime = datetime.datetime.now().strftime("%B %d, %Y - I:%M%p")

print("Running script")
st = time.time()
os.system("kernprof -l test_unittest.py")
execTime = time.time() - st
print("Execute time = %fs" % execTime)

profile_result = (
    subprocess.check_output(
        "python -m line_profiler test_unittest.py.lprof", shell=True
    )
    .decode("utf-8")
    .replace("\r\n", "\n")
)
print(profile_result)

with open("lprof_log.txt", "a") as lprof_fp:
    lprof_fp.write(
        "[%s]: Total execution time %.4fs\n%s\n\n\n"
        % (curTime, execTime, profile_result)
    )



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from typing import TYPE_CHECKING, NoReturn

from typing_extensions import Self

from ... import utils as ut
from ...localize import _
from ..allocator import ConstExpr, IsConstExpr
from .consttype import Byte, Dword, Word

if TYPE_CHECKING:
    from ..allocator.payload import ObjCollector, RlocInt_C, _PayloadBuffer
    from .rawtriggerdef import RawTrigger

_condtypes: dict[int, str] = {
    0: "(no condition)",
    1: "CountdownTimer",
    2: "Command",
    3: "Bring",
    4: "Accumulate",
    5: "Kills",
    6: "CommandMost",
    7: "CommandMostAt",
    8: "MostKills",
    9: "HighestScore",
    10: "MostResources",
    11: "Switch",
    12: "ElapsedTime",
    14: "Opponents",
    15: "Deaths",
    16: "CommandLeast",
    17: "CommandLeastAt",
    18: "LeastKills",
    19: "LowestScore",
    20: "LeastResources",
    21: "Score",
    22: "Always",
    23: "Never",
}


class Condition(ConstExpr):
    """Condition class.

    Memory layout:

     ======  =============  ========  ===========
     Offset  Field name     Position  EPD Player
     ======  =============  ========  ===========
       +00   locid           dword0   EPD(cond)+0
       +04   player          dword1   EPD(cond)+1
       +08   amount          dword2   EPD(cond)+2
       +0C   unitid          dword3   EPD(cond)+3
       +0E   comparison
       +0F   condtype
       +10   restype         dword4   EPD(cond)+4
       +11   flags
       +12   internal[2]
     ======  =============  ========  ===========
    """

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(
        self,
        locid: Dword,
        player: Dword,
        amount: Dword,
        unitid: Word,
        comparison: Byte,
        condtype: Byte,
        restype: Byte,
        flags: Byte,
        *,
        eudx: Word = 0,
    ) -> None:
        """See :mod:`eudplib.base.stockcond` for stock conditions list."""
        super().__init__()
        self.fields: list[Dword] = [
            locid,
            player,
            amount,
            unitid,
            comparison,
            condtype,
            restype,
            flags,
            eudx,
        ]
        self.parenttrg: RawTrigger | None = None
        self.condindex: int | None = None

    def __copy__(self) -> Condition:
        return self.__class__(*self.fields[:8], eudx=self.fields[8])  # type: ignore[arg-type]

    def disable(self) -> None:
        if isinstance(self.fields[7], ConstExpr):
            raise ut.EPError(_("Can't disable non-const Condition flags"))
        self.fields[7] |= 2

    # -------

    def _invalid_condition(self, i: int) -> str:
        condtype = self.fields[5]
        condname = _condtypes[condtype] if isinstance(condtype, int) else condtype
        return _("Invalid fields for condition{} {}:").format(i, condname)

    def CheckArgs(self, i: int) -> None:  # noqa: N802
        fields = self.fields

        if (
            IsConstExpr(fields[0])
            and IsConstExpr(fields[1])
            and IsConstExpr(fields[2])
            and isinstance(fields[3], int)
            and isinstance(fields[4], int)
            and isinstance(fields[5], int)
            and isinstance(fields[6], int)
            and isinstance(fields[7], int)
            and isinstance(fields[8], int)
        ):
            return

        error = [self._invalid_condition(i)]
        fieldname = [
            "location",
            "player",
            "amount",
            "unit_type",
            "comparison",
            "condition_type",
            "resource_type",
            "flags",
            "maskflag",
        ]
        condtype = fields[5]
        if isinstance(condtype, int):
            if condtype == 15:
                fieldname[0] = "bitmask"
            elif condtype == 11:
                fieldname[4] = "switchstate"
                fieldname[6] = "switchid"
            if condtype in (9, 19, 21):
                fieldname[6] = "scoretype"

        for i, field in enumerate(fields):
            if (i < 3 and not IsConstExpr(field)) or (
                i >= 3 and not isinstance(field, int)
            ):
                error.append(
                    "\t" + _("invalid {}: {}").format(fieldname[i], repr(field))
                )

        raise ut.EPError("\n".join(error))

    def SetParentTrigger(self, trg: RawTrigger, index: int) -> None:  # noqa: N802
        if self.parenttrg is not None:
            raise ut.EPError(_("Condition cannot be shared by two triggers."))
        if trg is None:
            raise ut.EPError(_("Trigger should not be null."))
        if not (0 <= index < 16):
            raise ut.EPError(_("Condition index should be 0 to 15"))

        self.parenttrg = trg
        self.condindex = index

    def Evaluate(self) -> RlocInt_C:  # noqa: N802
        if self.parenttrg is None or self.condindex is None:
            # fmt: off
            raise ut.EPError(_("Orphan condition. This often happens when you try to do arithmetics with conditions."))  # noqa: E501
            # fmt: on
        return self.parenttrg.Evaluate() + 8 + self.condindex * 20

    def CollectDependency(self, pbuffer: ObjCollector) -> None:  # noqa: N802
        for field in self.fields[:3]:
            pbuffer.WriteDword(field)  # type: ignore[arg-type]

    def WritePayload(self, pbuffer: _PayloadBuffer) -> None:  # noqa: N802
        pbuffer.WritePack("IIIHBBBBH", self.fields)  # type: ignore[arg-type]

    def __bool__(self) -> NoReturn:
        raise RuntimeError(_("To prevent error, Condition can't be put into if."))

    def negate(self) -> None:
        condtype = self.fields[5]
        ut.ep_assert(
            isinstance(condtype, int), f"Invalid condition type: {condtype}"
        )
        comparison_set = (1, 2, 3, 4, 5, 12, 14, 15, 21)
        always_set = (0, 22)
        never_set = (13, 23)
        if condtype in always_set:
            self.fields[5] = 23
        elif condtype in never_set:
            self.fields[5] = 22
        elif condtype == 11:  # Switch
            self.fields[4] ^= 1  # type: ignore[operator]
        elif condtype in comparison_set:
            bring_or_command = (2, 3)
            comparison = self.fields[4]
            amount = self.fields[2]
            ut.ep_assert(
                isinstance(comparison, int), f"Invalid comparison: {comparison}"
            )
            ut.ep_assert(isinstance(amount, int), f"Invalid amount: {amount}")
            amount &= 0xFFFFFFFF  # type: ignore[operator]
            if comparison == 10 and amount == 0:
                self.fields[4] = 0
                self.fields[2] = 1
            elif comparison == 0 and amount == 1:
                self.fields[4] = 10
                self.fields[2] = 0
            elif comparison == 0 and amount == 0:
                self.fields[5] = 23
            # AtMost and Exactly/AtLeast behaves differently in Bring/Command
            # (AtMost counts buildings on construction and does not count Egg)
            # So only exchanging (Exactly, 0) <-> (AtLeast, 1) is sound.
            #
            # See: https://cafe.naver.com/edac/book5095361/96809
            elif condtype in bring_or_command:
                # fmt: off
                raise ut.EPError(_("Bring and Command can't exchange AtMost and Exactly/AtLeast"))  # noqa: E501
                # fmt: on
            elif comparison in (0, 1):
                self.fields[4] ^= 1  # type: ignore[operator]
                self.fields[2] += -((-1) ** comparison)  # type: ignore[operator]
            elif comparison != 10:
                raise ut.EPError(
                    _('Invalid comparison "{}" in trigger index {}').format(
                        comparison, 0
                    )
                )
            elif condtype == 15 and self.fields[8] == ut.b2i2(ut.u2b("SC")):
                mask = self.fields[0] & 0xFFFFFFFF  # type: ignore[operator]
                if amount & (~mask):  # never
                    self.fields[5] = 23
                elif amount == mask:
                    self.fields[4] = 1
                    self.fields[2] = mask - 1
                else:
                    raise ut.EPError(_("Can't negate EUDX condition"))
            elif amount == 0xFFFFFFFF:
                self.fields[4] = 1  # AtMost
                self.fields[2] = 0xFFFFFFFE
            else:
                raise ut.EPError(_("Can't negate condition with comparison"))
        else:
            raise ut.EPError(_("Can't negate condition"))



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import Literal

from typing_extensions import Self

from .. import core as c


class EUDJumpBuffer(c.EUDObject):
    """Jump table buffer

    20 bytes per nextptr.
    """

    def __init__(self) -> None:
        super().__init__()

        self._jdict: dict[JumpTriggerForward, c.ConstExpr] = {}
        self._nextptrs: list[int | c.ConstExpr] = []

    def DynamicConstructed(self) -> Literal[True]:  # noqa: N802
        return True

    def Evaluate(self) -> c.RlocInt_C:  # noqa: N802
        return c.GetObjectAddr(self) - 4

    def create_jump_trigger(self, v, nextptr) -> c.ConstExpr:
        ret = self + 20 * len(self._nextptrs)
        self._nextptrs.append(nextptr)
        self._jdict[v] = ret
        return ret

    def create_jump_triggers(self, v, nextptrs) -> c.ConstExpr:
        ret = self + 20 * len(self._nextptrs)
        self._nextptrs.extend(nextptrs)
        self._jdict[v] = ret
        return ret

    def GetDataSize(self) -> int:  # noqa: N802
        return 2356 + 20 * len(self._nextptrs)

    def CollectDependency(self, emitbuffer) -> None:  # noqa: N802
        for nextptr in self._nextptrs:
            if not isinstance(nextptr, int):
                emitbuffer.WriteDword(nextptr)

    def WritePayload(self, emitbuffer) -> None:  # noqa: N802
        for nextptr in self._nextptrs[:17]:
            emitbuffer.WriteDword(nextptr)  # nextptr
            emitbuffer.WriteSpace(12)
            emitbuffer.WriteDword(0)  # nocond
        for nextptr in self._nextptrs[17:118]:
            emitbuffer.WriteDword(nextptr)  # nextptr
            emitbuffer.WriteSpace(4)
            emitbuffer.WriteDword(0)  # noact
            emitbuffer.WriteSpace(4)
            emitbuffer.WriteDword(0)  # nocond
        for nextptr in self._nextptrs[118:]:
            emitbuffer.WriteDword(nextptr)  # nextptr
            emitbuffer.WriteSpace(4)
            emitbuffer.WriteDword(0)  # noact
            emitbuffer.WriteDword(8)  # flags
            emitbuffer.WriteDword(0)  # nocond
        emitbuffer.WriteSpace(8)
        emitbuffer.WriteDword(0)  # noact
        emitbuffer.WriteDword(8)  # flags
        for _ in range(16):
            emitbuffer.WriteSpace(12)
            emitbuffer.WriteDword(0)  # noact
            emitbuffer.WriteDword(8)  # flags
        for _ in range(117 - 16):
            emitbuffer.WriteSpace(16)
            emitbuffer.WriteDword(8)  # flags


_jtb = None


def register_new_jumpbuffer():
    global _jtb
    _jtb = EUDJumpBuffer()


def get_current_jumpbuffer():
    return _jtb


c.RegisterCreatePayloadCallback(register_new_jumpbuffer)


# Unused jump table don't need to be allocated.
class JumpTriggerForward(c.ConstExpr):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(self, nextptrs) -> None:
        super().__init__()
        self._nextptrs = nextptrs

    def Evaluate(self):  # noqa: N802
        jtb = get_current_jumpbuffer()
        try:
            return jtb._jdict[self].Evaluate()
        except KeyError:
            jt = jtb.create_jump_triggers(self, self._nextptrs)
            return jt.Evaluate()



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from dataclasses import dataclass


@dataclass
class _UnlimiterBool:
    is_unlimiter_on: bool = False


_unlimiter: _UnlimiterBool = _UnlimiterBool(is_unlimiter_on=False)


def _turnUnlimiterOn() -> None:  # noqa: N802
    global _unlimiter
    _unlimiter.is_unlimiter_on = True


def IsUnlimiterOn() -> bool:  # noqa: N802
    global _unlimiter
    return _unlimiter.is_unlimiter_on is True



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import core as c
from .. import memio
from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.typehint import _Upgrade
from ..localize import _
from .offsetmap import (
    ByteMember,
    EPDOffsetMap,
    IconMember,
    RaceResearchMember,
    StatTextMember,
    WordMember,
)
from .player import TrgPlayer


class Upgrade(EPDOffsetMap, ConstType):
    __slots__ = ()
    mineralCostBase: ClassVar[WordMember] = WordMember("array", 0x655740)
    mineralCostFactor: ClassVar[WordMember] = WordMember("array", 0x6559C0)
    gasCostBase: ClassVar[WordMember] = WordMember("array", 0x655840)
    gasCostFactor: ClassVar[WordMember] = WordMember("array", 0x6557C0)
    timeCostBase: ClassVar[WordMember] = WordMember("array", 0x655B80)
    timeCostFactor: ClassVar[WordMember] = WordMember("array", 0x655940)
    requirementOffset: ClassVar[WordMember] = WordMember("array", 0x6558C0)
    icon: ClassVar[IconMember] = IconMember("array", 0x655AC0)
    label: ClassVar[StatTextMember] = StatTextMember("array", 0x655A40)
    race: ClassVar[RaceResearchMember] = RaceResearchMember("array", 0x655BFC)
    maxLevel: ClassVar[ByteMember] = ByteMember("array", 0x655700)
    broodWarFlag: ClassVar[ByteMember] = ByteMember("array", 0x655B3C)  # bool?

    @ut.classproperty
    def range(self):
        return (0, 60, 1)

    @classmethod
    def cast(cls, _from: _Upgrade):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _Upgrade) -> None:
        super().__init__(c.EncodeUpgrade(initval))

    def _current_upgrade_epd_subp(self, player: TrgPlayer | int):
        r: c.EUDVariable
        sc_researched = 0x58D2B0
        bw_researched = 0x58F32C
        u = self._value
        if isinstance(player, TrgPlayer):
            p = player._value
        else:
            p = player
        if isinstance(p, int) or isinstance(player, int):
            ut.ep_assert(
                0 <= p <= 11,
                _("To get current upgrade, player must be in range from P1 to P12"),
            )
            if isinstance(u, c.EUDVariable):
                # (case 1) player is const, upgrade is var
                q, r, update = self._get_stride_cache(1)  # type: ignore[assignment]
                update_start, update_restore, update_end = update
                branch, sc_branch, bw_branch = (c.Forward() for _ in range(3))
                epd, subp = c.EUDCreateVariables(2)
                # Trigger 1: update upgrade cache and initialize epd, subp
                c.RawTrigger(
                    nextptr=update_start,
                    actions=[
                        c.SetNextPtr(update_start, update_restore),
                        c.SetMemory(update_start + 348, c.SetTo, q.GetVTable()),
                        c.SetNextPtr(update_end, q.GetVTable()),
                        c.SetNextPtr(q.GetVTable(), r.GetVTable()),
                        c.SetNextPtr(r.GetVTable(), branch),
                        q.QueueAssignTo(epd),
                        r.QueueAssignTo(subp),
                    ],
                )
                # Trigger 2: conditional branch for SC and BW upgrades
                branch << c.RawTrigger(
                    nextptr=sc_branch,
                    conditions=u.AtLeast(46),
                    actions=c.SetNextPtr(branch, bw_branch),
                )
                # (SC) Trigger 3: add to epd, subp
                sc_branch << c.NextTrigger()
                quotient, remainder = divmod(p * 46, 4)
                c.SetVariables(
                    [epd, subp],
                    [ut.EPD(sc_researched) + quotient, remainder],
                    [c.Add, c.Add],
                )
                # (SC) Trigger 4: carry 4 subp to 1 epd
                # max(r) = 3, max(remainder) = 2, max(subp) = 5
                end = c.Forward()
                c.RawTrigger(
                    conditions=subp.AtLeast(4),
                    actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
                )
                c.SetNextTrigger(end)
                # (BW) Trigger 3: add to epd, subp (subp can be negative)
                bw_branch << c.NextTrigger()
                quotient, remainder = divmod(p * 15, 4)
                q_off, r_off = divmod(46, 4)
                c.SetVariables(
                    [epd, subp],
                    [
                        ut.EPD(bw_researched) + quotient - q_off,
                        remainder - r_off,
                    ],
                    [c.Add, c.Add],
                )
                # (BW) Trigger 4: nagative subp borrow from epd
                c.RawTrigger(
                    conditions=subp.AtLeast(0x80000000),
                    actions=[subp.AddNumber(4), epd.AddNumber(-1)],
                )
                # (BW) Trigger 5: carry 4 subp to 1 epd
                # max(r) = 3, max(remainder) = 3, subp -= 2, max(subp) = 4
                c.RawTrigger(
                    conditions=subp.AtLeast(4),
                    actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
                )
                end << c.NextTrigger()
            # (case 2) player is const, upgrade is const
            elif u < 46:
                quotient, remainder = divmod(sc_researched + u + p * 46, 4)
                return quotient + ut.EPD(0), remainder
            else:
                quotient, remainder = divmod(bw_researched + u - 46 + p * 15, 4)
                return quotient + ut.EPD(0), remainder

        elif isinstance(u, c.EUDVariable):
            # (case 3) player is var, upgrade is var
            q_upgrade, r_upgrade, update_upgrade = self._get_stride_cache(1)
            if not isinstance(r_upgrade, c.EUDVariable):
                assert False
            u_update_start, u_update_restore, u_update_end = update_upgrade
            q46, r46, update_player = player._get_stride_cache(46)
            q15, r15, _update = player._get_stride_cache(15)
            p_update_start, p_update_restore, p_update_end = update_player

            # Trigger 1: update upgrade cache
            update_p, upgrade_branch = c.Forward(), c.Forward()
            c.RawTrigger(
                nextptr=u_update_start,
                actions=[
                    c.SetNextPtr(u_update_start, u_update_restore),
                    c.SetMemory(u_update_start + 348, c.SetTo, update_p),
                    c.SetNextPtr(u_update_end, update_p),
                ],
            )
            # Trigger 2: update player cache and initialize epd, subp
            # Can't update both upgrade and player at the same time
            # because u_update_end and p_update_end can be aliased.
            sc_branch, bw_branch = c.Forward(), c.Forward()
            epd, subp = c.EUDCreateVariables(2)
            update_p << c.VProc(
                [q_upgrade, r_upgrade],
                [
                    q_upgrade.QueueAssignTo(epd),
                    r_upgrade.QueueAssignTo(subp),
                    c.SetNextPtr(p_update_start, p_update_restore),
                    c.SetMemory(p_update_start + 348, c.SetTo, upgrade_branch),
                    c.SetNextPtr(p_update_end, upgrade_branch),
                    # prepare conditional branch
                    c.SetNextPtr(upgrade_branch, sc_branch),
                ],
            )
            c.SetNextTrigger(p_update_start)
            # Trigger 3: conditional branch for SC and BW upgrades
            upgrade_branch << c.RawTrigger(
                nextptr=sc_branch,
                conditions=u.AtLeast(46),
                actions=c.SetNextPtr(upgrade_branch, bw_branch),
            )
            # (SC) Trigger 4: add to epd, subp
            sc_branch << c.NextTrigger()
            c.SetVariables(
                [epd, epd, subp],
                [ut.EPD(sc_researched), q46, r46],
                [c.Add, c.Add, c.Add],
            )
            # (SC) Trigger 5: carry 4 subp to 1 epd
            # max(r_upgrade) = 3, max(r46) = 2, max(subp) = 5
            end = c.Forward()
            c.RawTrigger(
                conditions=subp.AtLeast(4),
                actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
            )
            c.SetNextTrigger(end)
            # (BW) Trigger 4: add to epd, subp (subp can be negative)
            bw_branch << c.NextTrigger()
            q_off, r_off = divmod(46, 4)
            c.SetVariables(
                [epd, subp, epd, subp],
                [ut.EPD(bw_researched) - q_off, -r_off, q15, r15],
                [c.Add, c.Add, c.Add, c.Add],
            )
            # (BW) Trigger 5: nagative subp borrow from epd
            c.RawTrigger(
                conditions=subp.AtLeast(0x80000000),
                actions=[subp.AddNumber(4), epd.AddNumber(-1)],
            )
            # (BW) Trigger 6: carry 4 subp to 1 epd
            # max(r_upgrade) = 3, max(r15) = 3, subp -= 2, max(subp) = 4
            c.RawTrigger(
                conditions=subp.AtLeast(4),
                actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
            )
            end << c.NextTrigger()
        else:
            # (case 4) player is var, upgrade is const
            if u < 46:
                q, r, update = player._get_stride_cache(46)  # type: ignore[assignment]
                quotient, remainder = divmod(u, 4)
            else:
                q, r, update = player._get_stride_cache(15)  # type: ignore[assignment]
                quotient, remainder = divmod(u - 46, 4)
            update_start, update_restore, update_end = update
            nexttrg = c.Forward()
            c.RawTrigger(
                nextptr=update_start,
                actions=[
                    c.SetNextPtr(update_start, update_restore),
                    c.SetMemory(update_start + 348, c.SetTo, nexttrg),
                    c.SetNextPtr(update_end, nexttrg),
                ],
            )
            nexttrg << c.NextTrigger()
            if u < 46:
                epd = ut.EPD(sc_researched) + quotient + q
            else:
                epd = ut.EPD(bw_researched) + quotient + q
            subp = remainder + r if remainder else r
            # carry 4 subp -> 1 epd
            if u < 46 and remainder >= 2:
                c.RawTrigger(
                    conditions=subp.AtLeast(4),
                    actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
                )
            elif u >= 46 and remainder >= 1:
                c.RawTrigger(
                    conditions=subp.AtLeast(4),
                    actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
                )

        return epd, subp

    def __getitem__(self, player: TrgPlayer | int) -> c.EUDVariable:
        epd, subp = self._current_upgrade_epd_subp(player)
        return memio.f_bread_epd(epd, subp)

    def __setitem__(self, player: TrgPlayer | int, level) -> None:
        epd, subp = self._current_upgrade_epd_subp(player)
        memio.f_bwrite_epd(epd, subp, level)



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools
from typing import ClassVar

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...collections.eudarray import EUDArray
from ...maprw.injector.mainloop import (
    eud_onstart2,
    has_already_started,
)
from ...memio import (
    f_dwread_cp,
    f_epdread_epd,
    f_maskread_cp,
    f_repmovsd_epd,
    f_setcurpl2cpcache,
)
from ...scdata import CurrentPlayer, TrgUnit
from . import wiredata as wd

is64bit = c.EUDLightBool()
tranwire, grpwire, wirefram = c.EUDCreateVariables(3)
defaultwire = {}


def tranwire_default32():
    if "TranWire32" not in defaultwire:
        defaultwire["TranWire32"] = EUDArray(wd.TranWire32)
    return defaultwire["TranWire32"], wd.TranWire32


def grpwire_default32():
    if "GrpWire32" not in defaultwire:
        defaultwire["GrpWire32"] = EUDArray(wd.GrpWire32)
    return defaultwire["GrpWire32"], wd.GrpWire32


def wirefram_default32():
    if "Wirefram32" not in defaultwire:
        defaultwire["Wirefram32"] = EUDArray(wd.Wirefram32)
    return defaultwire["Wirefram32"], wd.Wirefram32


def tranwire_default64():
    if "TranWire64" not in defaultwire:
        defaultwire["TranWire64"] = EUDArray(wd.TranWire64)
    return defaultwire["TranWire64"], wd.TranWire64


def grpwire_default64():
    if "GrpWire64" not in defaultwire:
        defaultwire["GrpWire64"] = EUDArray(wd.GrpWire64)
    return defaultwire["GrpWire64"], wd.GrpWire64


def wirefram_default64():
    if "Wirefram64" not in defaultwire:
        defaultwire["Wirefram64"] = EUDArray(wd.Wirefram64)
    return defaultwire["Wirefram64"], wd.Wirefram64


class InitialWireframe:
    _collected: bool = False
    _tranwires: ClassVar[dict[int, int]] = {}
    _grpwires: ClassVar[dict[int, int]] = {}
    _wireframs: ClassVar[dict[int, int]] = {}

    @classmethod
    def _init(cls):
        global is64bit, tranwire, grpwire, wirefram
        f_epdread_epd(ut.EPD(0x68C1F4), ret=[tranwire])
        f_epdread_epd(ut.EPD(0x68C1FC), ret=[grpwire])
        f_epdread_epd(ut.EPD(0x68C204), ret=[wirefram])
        check_64bit = c.MemoryEPD(0, c.Exactly, 0)
        c.VProc(
            wirefram,
            [
                tranwire.AddNumber(1),
                grpwire.AddNumber(1),
                wirefram.AddNumber(1),
                c.SetMemory(check_64bit + 4, c.SetTo, 2),
                wirefram.QueueAddTo(ut.EPD(check_64bit) + 1),
            ],
        )
        if cs.EUDIf()(check_64bit):
            cs.DoActions(is64bit.Set())

            def create_init64(src, default, size):
                if src:
                    data = default[1]
                    init = [ut.i2b2(data[0])]
                    for n in range(size):
                        x = 2 * src[n] if n in src else 2 * n
                        if x == 2 * n:
                            init.append(bytes(8))
                        else:
                            init.append(ut.i2b2(data[x] >> 16))
                            init.append(ut.i2b4(data[x + 1]))
                            init.append(ut.i2b2(data[x + 2]))
                    init.append(ut.i2b2(data[-2] >> 16))
                    init.append(ut.i2b4(data[-1]))
                    init = c.Db(b"".join(init))
                    ut.ep_assert(
                        len(init.content) == (size + 1) * 8,
                        f"Size mismatch: {len(init.content)}",
                    )
                else:
                    init = default[0]
                return init if init._is_epd() else ut.EPD(init)

            tranwire_init = create_init64(cls._tranwires, tranwire_default64(), 106)
            grpwire_init = create_init64(cls._grpwires, grpwire_default64(), 131)
            wirefram_init = create_init64(cls._wireframs, wirefram_default64(), 228)

            f_repmovsd_epd(tranwire, tranwire_init, len(wd.TranWire64))
            f_repmovsd_epd(grpwire, grpwire_init, len(wd.GrpWire64))
            f_repmovsd_epd(wirefram, wirefram_init, len(wd.Wirefram64))
        if cs.EUDElse()():

            def init32(ptr, src, data):
                if not src:
                    return
                key_min = min(min(src.keys()), 5)  # Terran Siege Tank (Tank Mode)
                key_max = max(max(src.keys()), 87)  # Aldaris (Templar)

                def wf(n):
                    if n in src:
                        return 2 * src[n]
                    else:
                        return 2 * n

                init = [ut.i2b2(data[wf(key_min)])]
                for n in range(key_min, key_max + 1):
                    x = wf(n)
                    if x == 2 * n:
                        init.append(bytes(8))
                    else:
                        init.append(ut.i2b2(data[x] >> 16))
                        init.append(ut.i2b4(data[x + 1]))
                        init.append(ut.i2b2(data[x + 2]))
                init.append(ut.i2b2(data[x + 2] >> 16))
                init = c.Db(b"".join(init))
                ut.ep_assert(
                    len(init.content) == (key_max - key_min + 1) * 8 + 4,
                    f"Size mismatch: {len(init.content)}",
                )
                if key_min > 0:
                    ptr = ptr + 2 * key_min
                f_repmovsd_epd(ptr, ut.EPD(init), (key_max - key_min + 1) * 2 + 1)

            init32(tranwire, cls._tranwires, wd.TranWire32)
            init32(grpwire, cls._grpwires, wd.GrpWire32)
            init32(wirefram, cls._wireframs, wd.Wirefram32)
        cs.EUDEndIf()

    @classmethod
    def init(cls):
        ut.ep_assert(
            not has_already_started(),
            "Can't use EUDOnStart here. See https://cafe.naver.com/edac/69262",
        )
        if not cls._collected:
            cls._collected = True
            eud_onstart2(cls._init)

    @classmethod
    def _add(cls, unit, wireframe, length, wiredict):
        unit, wireframe = c.EncodeUnit(unit), c.EncodeUnit(wireframe)
        ut.ep_assert(
            isinstance(unit, int) and isinstance(wireframe, int),
            f"{unit} and {wireframe} should be valid unit",
        )
        if 0 <= unit < length and 0 <= wireframe < length:
            wiredict[unit] = wireframe
            cls.init()

    @classmethod
    def wireframes(cls, unit, wireframe):
        cls.tranwire(unit, wireframe)
        cls.grpwire(unit, wireframe)
        cls.wirefram(unit, wireframe)

    @classmethod
    def tranwire(cls, unit, wireframe):
        cls._add(unit, wireframe, 106, cls._tranwires)

    @classmethod
    def grpwire(cls, unit, wireframe):
        cls._add(unit, wireframe, 131, cls._grpwires)

    @classmethod
    def wirefram(cls, unit, wireframe):
        cls._add(unit, wireframe, 228, cls._wireframs)


def _set_wireframe(unit, wireframe, size, ptr, default32, default64):
    default32 = default32[0]
    default64 = default64[0]
    if not default32._is_epd():
        default32 = ut.EPD(default32)
    if not default64._is_epd():
        default64 = ut.EPD(default64)
    if cs.EUDIf()([unit <= size, wireframe <= size]):
        c.VProc(
            [unit, wireframe, ptr],
            [
                unit.QueueAddTo(unit),
                wireframe.QueueAddTo(wireframe),
                ptr.QueueAddTo(unit),
                c.SetMemory(0x6509B0, c.SetTo, default32),
            ],
        )
        c.RawTrigger(
            conditions=is64bit.IsSet(),
            actions=c.SetMemory(0x6509B0, c.SetTo, default64),
        )
        actions = [
            c.SetMemory(0x6509B0, c.SetTo, 0),
            c.SetDeathsX(CurrentPlayer, c.SetTo, 0, 0, 0xFFFF0000),
            c.SetMemory(0x6509B0, c.Add, 1),
            c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
            c.SetMemory(0x6509B0, c.Add, 1),
            c.SetDeathsX(CurrentPlayer, c.SetTo, 0, 0, 0xFFFF),
        ]
        c.VProc(
            [unit, wireframe],
            [
                unit.QueueAssignTo(ut.EPD(actions[0]) + 5),
                wireframe.SetDest(ut.EPD(0x6509B0)),
            ],
        )
        f_maskread_cp(0, 0xFFFF0000, ret=[ut.EPD(actions[1]) + 5])
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, 1))
        f_dwread_cp(0, ret=[ut.EPD(actions[3]) + 5])
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, 1))
        f_maskread_cp(0, 0xFFFF, ret=[ut.EPD(actions[5]) + 5])
        f_setcurpl2cpcache(actions=actions)
    cs.EUDEndIf()


@c.EUDTypedFunc([TrgUnit, TrgUnit])
def SetTranWire(unit, wireframe):  # noqa: N802
    InitialWireframe.init()
    _set_wireframe(
        unit, wireframe, 105, tranwire, tranwire_default32(), tranwire_default64()
    )


@c.EUDTypedFunc([TrgUnit, TrgUnit])
def SetGrpWire(unit, wireframe):  # noqa: N802
    InitialWireframe.init()
    _set_wireframe(
        unit, wireframe, 130, grpwire, grpwire_default32(), grpwire_default64()
    )


@c.EUDTypedFunc([TrgUnit, TrgUnit])
def SetWirefram(unit, wireframe):  # noqa: N802
    InitialWireframe.init()
    _set_wireframe(
        unit, wireframe, 227, wirefram, wirefram_default32(), wirefram_default64()
    )


@c.EUDTypedFunc([TrgUnit, TrgUnit])
def SetWireframes(unit, wireframe):  # noqa: N802
    InitialWireframe.init()
    SetTranWire(unit, wireframe)
    SetGrpWire(unit, wireframe)
    SetWirefram(unit, wireframe)


def Is64BitWireframe() -> c.Condition:  # noqa: N802
    """
    Check whether StarCraft is 32 bit or 64 bit, by inspecting wireframe.

    Local (desync) condition.
    Is64BitWireframe() loads custom wireframes,
    causing wireframe graphic glitches by StarCraft.
    """
    InitialWireframe.init()
    return is64bit.IsSet()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.eudfunc.eudf import _EUDPredefineParam, _EUDPredefineReturn
from . import dwepdio as dwm
from . import modcurpl as cp
from . import readtable


@_EUDPredefineReturn(2)
@c.EUDFunc
def _reader():
    ptr, epd = _reader._frets
    cs.DoActions(ptr.SetNumber(0), epd.SetNumber(ut.EPD(0)))
    for i in ut._rand_lst(range(32)):
        c.RawTrigger(
            conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, 2**i),
            actions=[
                ptr.AddNumber(2**i),
                epd.AddNumber(2 ** (i - 2)),
            ]
            if i >= 2
            else ptr.AddNumber(2**i),
        )

    # return ptr, epd


def f_dwepdread_cp(cpo, **kwargs):
    if not (isinstance(cpo, int) and cpo == 0):
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, cpo))
    ptr, epd = _reader(**kwargs)
    if not (isinstance(cpo, int) and cpo == 0):
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))
    return ptr, epd


def f_dwread_cp(cpo, *, ret=None):
    return readtable._cp_caller(readtable._get(0xFFFFFFFF, 0))(cpo, ret=ret)


def f_epdread_cp(cpo, *, ret=None):
    return readtable._cp_caller(
        readtable._get(0xFFFFFFFC, -2),
        _operations=[(ut.EPD(readtable.copy_ret) + 5, c.SetTo, ut.EPD(0))],
    )(cpo, ret=ret)


@_EUDPredefineReturn(1, 2)
@_EUDPredefineParam(1)
@c.EUDFunc
def _wreader(subp):
    w = _wreader._frets[0]
    cs.EUDSwitch(subp)
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    setnptr = c.SetNextPtr(readtable.read_end_common, block["swend"])
    setdest = c.SetMemory(readtable.copy_ret + 16, c.SetTo, ut.EPD(w.getValueAddr()))
    block["_actions"].extend([setnptr, setdest])
    for i in range(3):
        if cs.EUDSwitchCase()(i):
            readtrg = readtable._insert_or_get(0xFFFF << (8 * i), -(8 * i))
            c.SetNextTrigger(readtrg)

    if cs.EUDSwitchCase()(3):
        readtrg = readtable._insert_or_get(0xFF000000, -24)
        trg2, trg3 = c.Forward(), c.Forward()
        c.RawTrigger(
            nextptr=readtrg,
            actions=c.SetNextPtr(readtable.read_end_common, trg2),
        )
        readtrg = readtable._insert_or_get(0xFF, 8)
        trg2 << c.RawTrigger(
            nextptr=readtrg,
            actions=[
                c.SetMemory(0x6509B0, c.Add, 1),
                c.SetMemoryX(readtable.copy_ret + 24, c.SetTo, 8 << 24, 0xFF000000),
                c.SetNextPtr(readtable.read_end_common, trg3),
            ],
        )
        trg3 << c.RawTrigger(
            actions=[
                c.SetMemoryX(readtable.copy_ret + 24, c.SetTo, 7 << 24, 0xFF000000),
                c.SetMemory(0x6509B0, c.Add, -1),
            ]
        )
    cs.EUDEndSwitch()
    # return w


def f_wread_cp(cpo, subp, *, ret=None):
    if isinstance(subp, int) and 0 <= subp <= 2:
        return readtable._cp_caller(
            readtable._insert_or_get(0xFFFF << (8 * subp), -(8 * subp))
        )(cpo, ret=ret)

    if not (isinstance(cpo, int) and cpo == 0):
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, cpo))
    w = _wreader(subp, ret=ret)
    if not (isinstance(cpo, int) and cpo == 0):
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))
    return w


@_EUDPredefineReturn(1, 2)
@_EUDPredefineParam(1)
@c.EUDFunc
def _breader(subp):
    b = _breader._frets[0]
    cs.EUDSwitch(subp)
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    setnptr = c.SetNextPtr(readtable.read_end_common, block["swend"])
    setdest = c.SetMemory(readtable.copy_ret + 16, c.SetTo, ut.EPD(b.getValueAddr()))
    block["_actions"].extend([setnptr, setdest])
    for i in range(4):
        if cs.EUDSwitchCase()(i):
            readtrg = readtable._insert_or_get(0xFF << (8 * i), -(8 * i))
            c.SetNextTrigger(readtrg)
    cs.EUDEndSwitch()
    # return b


def f_bread_cp(cpo, subp, *, ret=None):
    if isinstance(subp, int) and 0 <= subp <= 3:
        return readtable._cp_caller(
            readtable._insert_or_get(0xFF << (8 * subp), -(8 * subp))
        )(cpo, ret=ret)

    if not (isinstance(cpo, int) and cpo == 0):
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, cpo))
    b = _breader(subp, rete=ret)
    if not (isinstance(cpo, int) and cpo == 0):
        cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))
    return b


def f_dwwrite_cp(cpo, value):
    if isinstance(cpo, int) and cpo == 0 and c.IsEUDVariable(value):
        c.VProc(value, value.QueueAssignTo(cp.CP))
    elif isinstance(cpo, int) and cpo == 0:
        cs.DoActions(c.SetDeaths(cp.CP, c.SetTo, value, 0))
    else:
        cs.DoActions(
            c.SetMemory(0x6509B0, c.Add, cpo),
            c.SetDeaths(cp.CP, c.SetTo, value, 0),
            c.SetMemory(0x6509B0, c.Add, -cpo),
        )


def f_dwadd_cp(cpo, value):
    if isinstance(cpo, int) and cpo == 0 and c.IsEUDVariable(value):
        c.VProc(value, value.QueueAddTo(cp.CP))
    elif isinstance(cpo, int) and cpo == 0:
        cs.DoActions(c.SetDeaths(cp.CP, c.Add, value, 0))
    else:
        cs.DoActions(
            c.SetMemory(0x6509B0, c.Add, cpo),
            c.SetDeaths(cp.CP, c.Add, value, 0),
            c.SetMemory(0x6509B0, c.Add, -cpo),
        )


def f_dwsubtract_cp(cpo, value):
    if isinstance(cpo, int) and cpo == 0 and c.IsEUDVariable(value):
        c.VProc(value, value.QueueSubtractTo(cp.CP))
    elif isinstance(cpo, int) and cpo == 0:
        cs.DoActions(c.SetDeaths(cp.CP, c.Subtract, value, 0))
    else:
        cs.DoActions(
            c.SetMemory(0x6509B0, c.Add, cpo),
            c.SetDeaths(cp.CP, c.Subtract, value, 0),
            c.SetMemory(0x6509B0, c.Add, -cpo),
        )


@c.EUDFunc
def _wwriter(subp, w):
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(3)):
        if cs.EUDSwitchCase()(i):
            c.RawTrigger(
                actions=c.SetDeathsX(cp.CP, c.SetTo, 0, 0, 65535 << (8 * i))
            )
            c.SeqCompute([(cp.CP, c.Add, w * (256**i))])
            cs.EUDBreak()

    # Things gets complicated on this case.
    # We won't hand-optimize this case. This is a very, very rare case
    if cs.EUDSwitchCase()(3):
        b0, b1 = dwm.f_dwbreak(w)[2:4]
        f_bwrite_cp(0, 3, b0)
        f_bwrite_cp(1, 0, b1)

    cs.EUDEndSwitch()


def f_wwrite_cp(cpo, subp, w):
    try:
        cs.DoActions(
            []
            if isinstance(cpo, int) and cpo == 0
            else c.SetMemory(0x6509B0, c.Add, cpo),
            c.SetDeathsX(
                cp.CP,
                c.SetTo,
                w * (256**subp),
                0,
                65535 << (8 * subp),
            ),
            []
            if isinstance(cpo, int) and cpo == 0
            else c.SetMemory(0x6509B0, c.Add, -cpo),
        )
    except TypeError:
        if not (isinstance(cpo, int) and cpo == 0):
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, cpo))
        _wwriter(subp, w)
        if not (isinstance(cpo, int) and cpo == 0):
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))


@c.EUDFunc
def _bwriter(subp, b):
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(4)):
        if cs.EUDSwitchCase()(i):
            c.RawTrigger(actions=c.SetDeathsX(cp.CP, c.SetTo, 0, 0, 255 << (8 * i)))

            c.SeqCompute([(cp.CP, c.Add, b * (256**i))])
            cs.EUDBreak()
    cs.EUDEndSwitch()
    return b


def f_bwrite_cp(cpo, subp, b):
    try:
        cs.DoActions(
            []
            if isinstance(cpo, int) and cpo == 0
            else c.SetMemory(0x6509B0, c.Add, cpo),
            c.SetDeathsX(
                cp.CP,
                c.SetTo,
                b * (256**subp),
                0,
                255 << (8 * subp),
            ),
            []
            if isinstance(cpo, int) and cpo == 0
            else c.SetMemory(0x6509B0, c.Add, -cpo),
        )
    except TypeError:
        if not (isinstance(cpo, int) and cpo == 0):
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, cpo))
        _bwriter(subp, b)
        if not (isinstance(cpo, int) and cpo == 0):
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))


def f_maskwrite_cp(cpo, value, mask):
    if isinstance(cpo, int) and cpo == 0:
        cs.DoActions(c.SetDeathsX(cp.CP, c.SetTo, value, 0, mask))
    else:
        cs.DoActions(
            c.SetMemory(0x6509B0, c.Add, cpo),
            c.SetDeathsX(cp.CP, c.SetTo, value, 0, mask),
            c.SetMemory(0x6509B0, c.Add, -cpo),
        )



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from os import urandom

from ... import core as c
from ...core.allocator.payload import _set_payload_logger_mode
from ...core.mapdata.chktok import CHK
from ...utils.blockstru import BlockStruManager, set_blockstru_manager
from .inj_finalizer import create_inject_finalizer
from .payload_init import initialize_payload
from .payload_reloc import create_payload_relocator
from .vector_reloc import create_vector_relocator

skip_payload_relocator = False


def PRT_SkipPayloadRelocator(enable: bool) -> None:  # noqa: N802
    global skip_payload_relocator
    skip_payload_relocator = enable


def apply_injector(chkt: CHK, root: c.Forward | c.RawTrigger) -> None:
    # Create injector triggers
    bsm = BlockStruManager()
    prev_bsm = set_blockstru_manager(bsm)
    if skip_payload_relocator:
        mrgndata = None
    else:
        mrgndata = urandom(5100)

    root = create_inject_finalizer(chkt, root, mrgndata)

    _set_payload_logger_mode(True)
    payload = c.CreatePayload(root)
    _set_payload_logger_mode(False)

    if skip_payload_relocator:
        payload = create_payload_relocator(payload)
        create_vector_relocator(chkt, payload)
    else:
        initialize_payload(chkt, payload, mrgndata)

    set_blockstru_manager(prev_bsm)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from . import modcurpl as cp


class CPByteWriter:
    """Write byte by byte"""

    def __init__(self):
        self._suboffset = c.EUDLightVariable()
        self._b = [c.EUDVariable(ut.b2i1(b"\r"))] + [
            c.EUDLightVariable(ut.b2i1(b"\r")) for _ in range(3)
        ]

    @c.EUDMethod
    def writebyte(self, byte):
        """Write byte to current position.

        Write a byte to current position of EUDByteWriter. Writer will advance
        by 1 byte.

        .. note::
            Bytes could be buffered before written to memory. After you
            finished using writebytes, you must call `flushdword` to flush the
            buffer.
        """
        cs.EUDSwitch(self._suboffset)
        for i in ut._rand_lst(range(3)):
            if cs.EUDSwitchCase()(i):
                c.VProc(
                    byte,
                    [byte.SetDest(self._b[i]), self._suboffset.AddNumber(1)],
                )
                cs.EUDBreak()

        if cs.EUDSwitchCase()(3):
            c.VProc(byte, byte.SetDest(self._b[3]))
            self.flushdword()

        cs.EUDEndSwitch()

    @c.EUDMethod
    def flushdword(self):
        """Flush buffer."""
        # mux bytes
        cs.DoActions(c.SetDeaths(cp.CP, c.SetTo, 0, 0))
        if cs.EUDIf()(self._suboffset == 0):
            c.EUDReturn()
        cs.EUDEndIf()

        c.VProc(self._b[0], self._b[0].SetDest(cp.CP))
        for k in ut._rand_lst(range(8, 32)):
            i, j = divmod(k, 8)
            c.RawTrigger(
                conditions=self._b[i].AtLeastX(1, 2**j),
                actions=c.SetDeaths(cp.CP, c.Add, 2**k, 0),
            )
        cs.DoActions(
            *c.AddCurrentPlayer(1),
            self._suboffset.SetNumber(0),
            self._b[0].SetNumber(b"\r"[0]),
            self._b[1].SetNumber(b"\r"[0]),
            self._b[2].SetNumber(b"\r"[0]),
            self._b[3].SetNumber(b"\r"[0]),
        )



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import utils as ut
from ..localize import _
from . import loopblock as lb
from . import swblock as sb


def EUDContinue() -> None:  # noqa: N802
    return lb.eudloop_continue()


def EUDContinueIf(conditions) -> None:  # noqa: N802
    return lb.eudloop_continue_if(conditions)


def EUDContinueIfNot(conditions) -> None:  # noqa: N802
    return lb.eudloop_continue_ifnot(conditions)


def EUDSetContinuePoint() -> None:  # noqa: N802
    return lb.set_continuepoint()


def EUDIsContinuePointSet() -> bool:  # noqa: N802
    return lb.is_continuepoint_set()


# -------


def EUDBreak() -> None:  # noqa: N802
    for block in reversed(ut.EUDGetBlockList()):
        if lb._is_loopblock(block[1]):
            lb.eudloop_break()
            return
        elif sb._is_switch_blockid(block[0]):
            sb.eudswitch_break()
            return

    raise ut.EPError(_("No loop/switch block surrounding this code area"))


def EUDBreakIf(conditions) -> None:  # noqa: N802
    for block in reversed(ut.EUDGetBlockList()):
        if lb._is_loopblock(block[1]):
            lb.eudloop_break_if(conditions)
            return
        elif sb._is_switch_blockid(block[0]):
            sb.eudswitch_break_if(conditions)
            return

    raise ut.EPError(_("No loop/switch block surrounding this code area"))


def EUDBreakIfNot(conditions) -> None:  # noqa: N802
    for block in reversed(ut.EUDGetBlockList()):
        if lb._is_loopblock(block[1]):
            lb.eudloop_break_ifnot(conditions)
            return
        elif sb._is_switch_blockid(block[0]):
            sb.eudswitch_break_ifnot(conditions)
            return

    raise ut.EPError(_("No loop/switch block surrounding this code area"))



import time
import helper

start_time = time.time()
DoCoverageTest = False

if DoCoverageTest:
    import coverage

    cov = coverage.Coverage()
    cov.start()


helper.LoadMap("outputmap/basemap/basemap_wireframe.scx")
helper.EP_SetRValueStrictMode(True)

with open("outputmap/basemap/RequireData", "rb") as reqfile:
    helper._add_datadumper(reqfile.read(), [0x58D740], set(["copy"]))

helper.CompressPayload(True)
helper.ShufflePayload(False)

# profile_tool.profile(f, "profile.json")
helper.SaveMap("outputmap/test_datadumper.scx", helper._testmain)
print("--- %s seconds ---" % (time.time() - start_time))

if DoCoverageTest:
    cov.stop()
    cov.html_report(include=["C:\\gitclones\\eudtrglib\\eudplib\\*"])
